# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ez_wsi_dicomweb.slide_level_map."""
import json
import random
from typing import Iterator, Mapping

from absl.testing import absltest
from absl.testing import parameterized
from ez_wsi_dicomweb import dicom_test_utils
from ez_wsi_dicomweb import dicom_web_interface
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import slide_level_map

from hcls_imaging_ml_toolkit import dicom_path
from hcls_imaging_ml_toolkit import tags


class SlideLevelMapTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.dicom_objects = []
    with open(dicom_test_utils.SAMPLE_INSTANCES_PATH) as json_file:
      data = json.load(json_file)
      data = data['test_data']
      for x in data:
        dicom_object = dicom_web_interface.DicomObject(
            dicom_path.FromString(x['path']), x['dicom_tags']
        )
        self.dicom_objects.append(dicom_object)
      # Shuffle the order of the DICOM objects.
      random.shuffle(self.dicom_objects)

  def test_level_map_property(self):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)

    self.assertIsInstance(level_map.level_map, Mapping)

  def test_constructor_with_normal_input(self):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    self.assertEqual(1, level_map.level_index_min)
    self.assertEqual(10, level_map.level_index_max)
    level = level_map.get_level(1)
    self.assertIsNotNone(level, 'The native level cannot be missing.')
    if level is not None:
      self.assertEqual(0, level.frame_number_min)
      self.assertEqual(79002, level.frame_number_max)

  def test_raises_if_instances_have_different_transfer_sytax(self):
    with open(
        'ez_wsi_dicomweb/testdata/error_multi_transfer_syntax.json'
    ) as json_file:
      data = json.load(json_file)
      data = data['test_data']
      dicom_objects = [
          dicom_web_interface.DicomObject(
              dicom_path.FromString(x['path']), x['dicom_tags']
          )
          for x in data
      ]
    with self.assertRaises(
        ez_wsi_errors.SlideLevelContainsInstancesWithDifferentTransferSyntaxUIDError
    ):
      slide_level_map.SlideLevelMap(dicom_objects)

  def test_constructor_with_invalid_input_raises_error(self):
    empty_object_dict = []
    with self.assertRaises(ez_wsi_errors.NoDicomLevelsDetectedError):
      slide_level_map.SlideLevelMap(empty_object_dict)

  def test_get_level_with_invalid_input_return_none(self):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    self.assertIsNone(
        level_map.get_level(0), 'Level 0 should not exist in the testing map.'
    )

  def test_get_level_with_valid_input(self):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(1)
    self.assertIsNotNone(level, 'Level 1 should exist in the testing map.')
    if level is not None:  # level is Optional[Level]
      self.assertEqual(level.width, 98816)
      self.assertEqual(level.height, 199168)
      self.assertEqual(level.samples_per_pixel, 3)
      self.assertEqual(level.bits_allocated, 8)
      self.assertEqual(level.high_bit, 7)
      self.assertAlmostEqual(level.pixel_spacing_x_mm, 0.00024309, 7)
      self.assertAlmostEqual(level.pixel_spacing_y_mm, 0.00024309, 7)
      self.assertEqual(level.frame_width, 500)
      self.assertEqual(level.frame_height, 500)
      self.assertLen(level.instances, 39)

  @parameterized.parameters(0.0001, 0.0026, 0.000001)
  def test_get_level_by_pixel_spacing_with_out_of_range_pixel_spacing(
      self, pixel_spacing: float
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level_by_pixel_spacing(pixel_spacing)
    self.assertIsNone(
        level,
        (
            'There should be no level that has pixel spacing close to '
            f'{pixel_spacing}. Actual: matched to '
            f'{level.pixel_spacing_x_mm if level is not None else None}'
        ),
    )

  @parameterized.parameters(
      (0.00024, 1), (0.00028, 1), (0.00049, 2), (0.00095, 3), (0.1244, 10)
  )
  def test_get_level_by_pixel_spacing_with_in_range_pixel_spacing(
      self, pixel_spacing: float, expected_level: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level_by_pixel_spacing(pixel_spacing)
    self.assertIsNotNone(
        level,
        (
            'There should have a level that has pixel spacing '
            f'close to {pixel_spacing}.'
        ),
    )
    self.assertEqual(
        level.level_index if level is not None else 0,
        expected_level,
        'The found level should have an index of {expected_level}.',
    )

  @parameterized.parameters((1, 0), (1, 2047), (1, 40960), (2, 2048))
  def test_get_instance_by_frame_with_existing_frames(
      self, level_index: int, frame_index: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      instance = level.get_instance_by_frame(frame_index)
      self.assertIsNotNone(
          instance,
          (
              'There should exist an instance containing frame: '
              f'{frame_index} at level {level_index}.'
          ),
      )

  @parameterized.parameters((1, -1), (1, 1000000), (10, 40960))
  def test_get_instance_by_frame_with_nonexisting_frames(
      self, level_index: int, frame_index: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      instance = level.get_instance_by_frame(frame_index)
      self.assertIsNone(
          instance,
          f'Frame {frame_index} should not exist at level {level_index}.',
      )

  @parameterized.parameters(
      (1, 0, 0), (1, 2047, 2047), (1, 0, 40960), (2, 1, 2048), (10, 192, 388)
  )
  def test_get_instance_by_point_with_valid_point(
      self, level_index: int, x: int, y: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      instance = level.get_instance_by_point(x, y)
      self.assertIsNotNone(
          instance,
          (
              'There should exist an instance containing point: ('
              f'{x, y}) at level {level_index}.'
          ),
      )

  @parameterized.parameters((1, -1, 0), (1, 0, 409600), (10, 193, 389))
  def test_get_instance_by_point_with_out_of_range_point(
      self, level_index: int, x: int, y: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      with self.assertRaises(ez_wsi_errors.CoordinateOutofImageDimensionsError):
        level.get_instance_by_point(x, y)

  @parameterized.parameters(
      (1, 0, 0, 0),
      (1, 2047, 2047, 796),
      (1, 0, 40960, 16038),
      (2, 1, 2048, 396),
      (10, 192, 388, 0),
  )
  def test_get_frame_by_point_with_valid_point(
      self, level_index: int, x: int, y: int, frame_number: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      calculated_frame_number = level.get_frame_number_by_point(x, y)
      self.assertEqual(
          frame_number,
          calculated_frame_number,
          (
              f'There calculated frame number({calculated_frame_number}) '
              f'does not match with expected value: {frame_number}'
          ),
      )

  @parameterized.parameters((1, -1, 0), (1, 0, 409600), (10, 193, 389))
  def test_get_frame_by_point_with_out_of_range_point(
      self, level_index: int, x: int, y: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      with self.assertRaises(ez_wsi_errors.CoordinateOutofImageDimensionsError):
        level.get_instance_by_point(x, y)

  @parameterized.parameters((1, -1), (1, 6553600), (10, 20))
  def test_get_frame_position_with_out_of_range_input(
      self, level_index: int, frame_number: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      with self.assertRaises(ez_wsi_errors.FrameNumberOutofBoundsError):
        level.get_frame_position(frame_number)

  @parameterized.parameters(
      (1, 0, 0, 0),
      (1, 7, 3500, 0),
      (1, 256, 29000, 500),
      (2, 5, 2500, 0),
      (10, 0, 0, 0),
  )
  def test_get_frame_position_with_valid_input(
      self, level_index: int, frame_number: int, x: int, y: int
  ):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(level_index)
    self.assertIsNotNone(
        level, f'There should exist a level at index {level_index}.'
    )
    if level is not None:
      pos_x, pos_y = level.get_frame_position(frame_number)
      self.assertEqual(
          (x, y),
          (pos_x, pos_y),
          (
              f'The returned corner ({pos_x, pos_y}) do not match expectation:'
              f' ({x, y}) at level {level_index}.'
          ),
      )

  @parameterized.parameters((0, 1), (1, 2), (2047, 2048), (2048, 1))
  def test_index_from_frame_number(self, frame_number: int, frame_index: int):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    level = level_map.get_level(1)
    self.assertIsNotNone(level)
    if level is not None:
      instance = level.get_instance_by_frame(frame_number)
      self.assertIsNotNone(instance)
      if instance is not None:
        self.assertEqual(
            frame_index, instance.frame_index_from_frame_number(frame_number)
        )

  def test_wsi_sop_class_id_detection(self):
    vl_micro_sop_class_id = "1.2.840.10008.5.1.4.1.1.77.1.6"
    self.assertEqual(vl_micro_sop_class_id, '1.2.840.10008.5.1.4.1.1.77.1.6')

  def test_slide_level_map_excludes_thumbnail_macro_and_label_images(self):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    for level in level_map._level_map.values():
      for instance in level.instances.values():
        image_type = instance.dicom_object.get_list_value(tags.IMAGE_TYPE)
        if image_type:
          # Validate that Thumbnail, label, and overview images are not being
          # stored.
          self.assertFalse(
              frozenset(image_type)
              & frozenset(('LABEL', 'THUMBNAIL', 'OVERVIEW'))
          )
        else:
          self.assertIsNone(image_type)

  def test_instance_iterator(self):
    level_map = slide_level_map.SlideLevelMap(self.dicom_objects)
    self.assertIsInstance(level_map._level_map[1].instance_iterator, Iterator)
    self.assertLen(list(level_map._level_map[1].instance_iterator), 39)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ez_wsi_dicomweb.ez_wsi_errors."""
import inspect
import sys

from absl.testing import absltest
from absl.testing import parameterized

from ez_wsi_dicomweb import ez_wsi_errors


class EZWsiErrorsTest(parameterized.TestCase):

  def test_all_errors_use_base_class(self):
    for _, cls in inspect.getmembers(
        sys.modules[ez_wsi_errors.__name__], inspect.isclass
    ):
      self.assertTrue(issubclass(cls, ez_wsi_errors.EZWsiError))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Place for common constants and methods accross tests for EZ WSI DicomWeb."""
import json
from unittest import mock

from ez_wsi_dicomweb import dicom_web_interface

from hcls_imaging_ml_toolkit import dicom_path

# The path to the testing DICOM series data.
TEST_SLIDE_UID_1 = '1.22.333.101:1.22.333.201'
TEST_STORE_PATH = (
    'projects/project_name/locations/us-west1/datasets/'
    'dataset_name/dicomStores/dicom_store_name'
)
# The DICOMStore path to a testing series.
TEST_DICOM_SERIES = (
    TEST_STORE_PATH + '/dicomWeb/studies/1.22.333.101/series/1.22.333.201'
)

TEST_DICOM_SERIES_2 = (
    TEST_STORE_PATH + '/dicomWeb/studies/1.22.333.101/series/1.22.333.999'
)
# Test DICOM object UIDs.
TEST_STUDY_UID_1 = '1.22.333.101'
TEST_STUDY_UID_2 = '1.22.333.102'
TEST_SERIES_UID_1 = '1.22.333.201'
TEST_SERIES_UID_2 = '1.22.333.202'
TEST_INSTANCE_UID_1 = '1.22.333.301'
TEST_INSTANCE_UID_2 = '1.22.333.302'
# The path to the testing DICOM series data.
TEST_DATA_PATH = 'ez_wsi_dicomweb/testdata'
SAMPLE_INSTANCES_PATH = TEST_DATA_PATH + '/sample_instances.json'
INSTANCE_CONCATENATION_TEST_DATA_PATH = (
    TEST_DATA_PATH + '/concatenation_instance_test.json'
)
TEST_JPEG_PATH = TEST_DATA_PATH + '/google.jpg'
TEST_DICOM_STORE_PATH = TEST_DATA_PATH + '/test.dcm'


def create_mock_dicom_web_interface(test_data_path: str):
  """Helper function to create a mocked DicomWebInterface object."""
  with open(test_data_path, 'rt') as json_file:
    data = json.load(json_file)
    data = data['test_data']
    dicom_objects = [
        dicom_web_interface.DicomObject(
            dicom_path.FromString(x['path']), x['dicom_tags']
        )
        for x in data
    ]
  mock_dwi = mock.create_autospec(dicom_web_interface.DicomWebInterface)
  mock_dwi.get_instances.return_value = dicom_objects
  return mock_dwi



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ez_wsi_dicomweb.dicom_slide."""
import collections
import dataclasses
import typing
from typing import Iterator, List, Optional, Tuple
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from ez_wsi_dicomweb import abstract_slide_frame_cache
from ez_wsi_dicomweb import dicom_slide
from ez_wsi_dicomweb import dicom_test_utils
from ez_wsi_dicomweb import dicom_web_interface
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import magnification
from ez_wsi_dicomweb import slide_level_map
import numpy as np
import PIL.Image

from hcls_imaging_ml_toolkit import dicom_path


def _init_test_slide_level_map(
    slide: dicom_slide.DicomSlide,
    width: int,
    height: int,
    frame_width: int,
    frame_height: int,
    frame_number_min: int,
    frame_number_max: int,
    samples_per_pixel: int,
    transfer_syntax_uid: str,
    mock_path: bool = False,
):
  new_map = collections.OrderedDict()
  if 1 not in slide._level_map._level_map:
    spacing_placeholder = 0.01
    instances = {}
    new_map[1] = slide_level_map.Level(
        1,
        width,
        height,
        samples_per_pixel,
        8,
        7,
        spacing_placeholder,
        spacing_placeholder,
        frame_width,
        frame_height,
        frame_number_min,
        frame_number_max,
        instances,
        transfer_syntax_uid,
    )
  else:
    for key, value in slide._level_map._level_map.items():
      if key == 1:
        new_map[key] = dataclasses.replace(
            slide._level_map._level_map[key],
            width=width,
            height=height,
            frame_width=frame_width,
            frame_height=frame_height,
            frame_number_min=frame_number_min,
            frame_number_max=frame_number_max,
            samples_per_pixel=samples_per_pixel,
            transfer_syntax_uid=transfer_syntax_uid,
        )
      else:
        new_map[key] = value
  count = 1
  updated_instances = collections.OrderedDict()
  if new_map[1].instances is not None:  # pytype: disable=attribute-error
    for key, instance in new_map[1].instances.items():  # pytype: disable=attribute-error
      if mock_path:
        dicom_object = dataclasses.replace(
            instance.dicom_object,
            path=dicom_path.FromPath(slide.path, instance_uid=str(count)),
        )
      else:
        dicom_object = instance.dicom_object
      updated_instances[key] = dataclasses.replace(
          instance,
          dicom_object=dicom_object,
      )
      count += 1
    new_map[1] = dataclasses.replace(new_map[1], instances=updated_instances)
  slide._level_map._level_map = new_map
  return slide


def _fake_get_frame_raw_image(
    unused_param: dicom_path.Path,
    x: int,
    transcode_frame: dicom_web_interface.TranscodeDicomFrame,
) -> bytes:
  """Mocks dicom_web_interface get_frame_image for uncompressed transfer syntax.

  uncompressed transfer syntax = 1.2.840.10008.1.2.1

  Args:
    unused_param: Path to DICOM instance
    x: Frame number.
    transcode_frame: How DICOM frame should be transcoded.

  Returns:
    Mocked image bytes.
  """
  assert (
      transcode_frame
      == dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN
  )
  return bytes(range((x - 1) * 4, x * 4))


class MockFrameCache:

  def is_supported_transfer_syntax(self, transfer_syntax: str) -> bool:
    return (
        transfer_syntax
        == dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN.value
    )

  def get_frame(
      self, instance_path: str, frame_count: int, frame_number: int
  ) -> Optional[bytes]:
    del instance_path, frame_count
    return _fake_get_frame_raw_image(
        mock.ANY,
        frame_number,
        dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN,
    )


def _fake_get_frame_image_jpeg(
    unused_param: dicom_path.Path,
    unused_x: int,
    transcode_frame: dicom_web_interface.TranscodeDicomFrame,
) -> bytes:
  """Mocks dicom_web_interface get_frame_image for jpeg transfer syntax.

  jpeg_transfer_syntax = 1.2.840.10008.1.2.4.50

  Args:
    unused_param: Path to DICOM instance
    unused_x: Frame number.
    transcode_frame: How DICOM frame should be transcoded.

  Returns:
    Mocked image bytes.
  """
  assert (
      transcode_frame
      == dicom_web_interface.TranscodeDicomFrame.DO_NOT_TRANSCODE
  )
  with open(dicom_test_utils.TEST_JPEG_PATH, 'rb') as infile:
    return infile.read()


class DicomSlideTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # Load DICOM objects for the testing slide.
    self.mock_dwi = dicom_test_utils.create_mock_dicom_web_interface(
        dicom_test_utils.SAMPLE_INSTANCES_PATH
    )
    self.dicom_series_path = dicom_path.FromString(
        dicom_test_utils.TEST_DICOM_SERIES
    )

  @parameterized.parameters([
      ('1.2.840.10008.1.2.4.50', True),  # JPEG Baseline
      ('1.2.840.10008.1.2.4.90', True),  # JPEG 2000 (Lossless)
      ('1.2.840.10008.1.2.4.91', True),  # JPEG 2000
      ('1.2.840.10008.1.2.1', True),  # Explicit VR Little Endian
      ('1.2.840.10008.1.2', True),  # 	Implicit VR Endian: Default
      ('1.2.3', False),
      ('1.2.840.10008.1.2.1.99', False),
      ('1.2.840.10008.1.2.2', False),
      ('1.2.840.10008.1.2.4.51', False),
      ('1.2.840.10008.1.2.4.70', False),
      ('1.2.840.10008.1.2.4.80', False),
      ('1.2.840.10008.1.2.4.81', False),
      ('1.2.840.10008.1.2.4.92', False),
      ('1.2.840.10008.1.2.4.93', False),
      ('1.2.840.10008.1.2.4.94', False),
      ('1.2.840.10008.1.2.4.95', False),
      ('1.2.840.10008.1.2.5', False),
      ('1.2.840.10008.1.2.6.1', False),
  ])
  def test_is_client_side_pixel_decoding_supported(
      self, val: str, expected: bool
  ):
    self.assertEqual(
        dicom_slide.is_client_side_pixel_decoding_supported(val), expected
    )

  def test_getstate(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )

    self.assertIn('_dwi', slide.__dict__)
    self.assertIn('_server_request_frame_cache', slide.__dict__)
    state = slide.__getstate__()
    self.assertNotIn('_dwi', state)
    self.assertNotIn('_server_request_frame_cache', state)

  def test_dicom_slide_equal(self):
    self.assertEqual(
        dicom_slide.DicomSlide(
            self.mock_dwi,
            self.dicom_series_path,
            enable_client_slide_frame_decompression=True,
        ),
        dicom_slide.DicomSlide(
            self.mock_dwi,
            self.dicom_series_path,
            enable_client_slide_frame_decompression=True,
        ),
    )

  def test_dicom_slide_not_equal_dicom_slide_object(self):
    self.assertNotEqual(
        dicom_slide.DicomSlide(
            self.mock_dwi,
            self.dicom_series_path,
            enable_client_slide_frame_decompression=True,
        ),
        dicom_slide.DicomSlide(
            self.mock_dwi,
            dicom_path.FromString(dicom_test_utils.TEST_DICOM_SERIES_2),
            enable_client_slide_frame_decompression=True,
        ),
    )

  def test_dicom_slide_not_equal_other_object(self):
    self.assertFalse(
        dicom_slide.DicomSlide(
            self.mock_dwi,
            self.dicom_series_path,
            enable_client_slide_frame_decompression=True,
        ).__eq__('foo')
    )

  def test_setstate(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    state = slide.__getstate__()
    origional_frame_cache_instance = slide._server_request_frame_cache
    self.assertIsNotNone(slide.dwi)

    slide.__setstate__(state)
    self.assertIsNone(slide.dwi)
    self.assertIsNot(
        slide._server_request_frame_cache, origional_frame_cache_instance
    )

  def test_dwi_setter_getter(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )

    val = mock.MagicMock()
    slide.dwi = val
    self.assertIs(slide.dwi, val)
    self.assertIs(slide._dwi, val)

  def test_get_magnifications(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    self.assertListEqual(
        slide.magnifications,
        [
            magnification.Magnification.FromString('40X'),
            magnification.Magnification.FromString('20X'),
            magnification.Magnification.FromString('10X'),
            magnification.Magnification.FromString('5X'),
            magnification.Magnification.FromString('2.5X'),
            magnification.Magnification.FromString('1.25X'),
            magnification.Magnification.FromString('0.625X'),
            magnification.Magnification.FromString('0.3125X'),
            magnification.Magnification.FromString('0.15625X'),
            magnification.Magnification.FromString('0.078125X'),
        ],
    )

  def test_constructor(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    self.assertIsNotNone(slide._level_map)
    self.assertEqual(
        '40X',
        slide.native_magnification.as_string,
        'The native mangification of the test slide must be 40X.',
    )
    self.assertEqual(199168, slide.total_pixel_matrix_rows)  # Tag: 00480007
    self.assertEqual(98816, slide.total_pixel_matrix_columns)  # Tag: 00480006
    self.assertEqual(
        np.uint8, slide.pixel_format, 'The pixel format is incorrect.'
    )

  def test_get_pixel_format_with_valid_input(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    level = slide.get_level_by_magnification(slide.native_magnification)
    self.assertEqual(
        np.uint8,
        dicom_slide._get_pixel_format(level),
        'The pixel format is not correct.',
    )

  def test_get_pixel_format_with_unsupported_pixel_format_raises_error(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    level = slide.get_level_by_magnification(slide.native_magnification)
    level = dataclasses.replace(level, bits_allocated=32)
    with self.assertRaises(ez_wsi_errors.UnsupportedPixelFormatError):
      dicom_slide._get_pixel_format(level)

  def test_get_native_level_with_valid_input(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    level = dicom_slide._get_native_level(slide._level_map)
    self.assertEqual(
        1, level.level_index, 'The index of the native level is incorrect.'
    )

  def test_get_native_level_with_missing_min_level_raises_error(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    slide._level_map._level_map[slide._level_map.level_index_min] = (
        None  # pytype: disable=unsupported-operands  # always-use-return-annotations
    )
    with self.assertRaises(ez_wsi_errors.MagnificationLevelNotFoundError):
      dicom_slide._get_native_level(slide._level_map)

  def test_get_level_with_nonexsting_mag_returns_none(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    level = slide.get_level_by_magnification(
        magnification.Magnification.FromString('80X')
    )
    self.assertIsNone(level, 'There should be no level at magnification 80X.')

  def test_get_level_with_existing_mag_returns_valid_level(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    level = slide.get_level_by_magnification(
        magnification.Magnification.FromString('10X')
    )
    self.assertIsNotNone(
        level, 'The testing slide should contain a level at magnification 10X.'
    )
    if level is not None:
      self.assertEqual(24704, level.width)
      self.assertEqual(49792, level.height)
      self.assertEqual(500, level.frame_width)
      self.assertEqual(500, level.frame_height)

  @parameterized.parameters(('40X', -1), ('40X', 655360), ('5X', 2047))
  def test_get_frame_with_out_of_range_frame_number_raise_error(
      self, mag: str, frame_number: int
  ):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    with self.assertRaises(ez_wsi_errors.InputFrameNumberOutOfRangeError):
      slide.get_frame(
          magnification.Magnification.FromString(mag),
          frame_number,
      )

  def test_get_frame_normal(self):
    self.mock_dwi.get_frame_image.return_value = b'abc123abc123'
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Set frame size to 2x2 for the first level.
    level_1 = slide._level_map._level_map[1]
    _init_test_slide_level_map(
        slide,
        level_1.width,
        level_1.height,
        2,
        2,
        level_1.frame_number_min,
        level_1.frame_number_max,
        level_1.samples_per_pixel,
        level_1.transfer_syntax_uid,
    )
    frame = slide.get_frame(slide.native_magnification, 1)
    self.assertIsNotNone(
        frame,
        (
            'The testing slide must contain a frame at index 1 on the native '
            'magnification level.'
        ),
    )
    if frame is not None:
      self.assertEqual(
          (2, 0, 2, 2),
          (frame.x_origin, frame.y_origin, frame.width, frame.height),
          'The location and dimension of the frame are not correct.',
      )
      self.assertTrue(
          np.array_equal(
              [[[97, 98, 99], [49, 50, 51]], [[97, 98, 99], [49, 50, 51]]],
              frame.image_np,
          ),
          'The returned frame is not as expected.',
      )

  @parameterized.named_parameters([
      dict(
          testcase_name='explicit_vr_little_endian',
          transfer_syntax_uid='1.2.840.10008.1.2.1',
          expected_mock_get_frame_from_cache_call_count=1,
          expected_frame_bytes=49,
      ),
      dict(
          testcase_name='implicit_vr_endian',
          transfer_syntax_uid='1.2.840.10008.1.2',
          expected_mock_get_frame_from_cache_call_count=1,
          expected_frame_bytes=49,
      ),
      dict(
          testcase_name='deflated_explicit_vr_little_endian',
          transfer_syntax_uid='1.2.840.10008.1.2.1.99',
          expected_mock_get_frame_from_cache_call_count=0,
          expected_frame_bytes=50,
      ),
      dict(
          testcase_name='explicit_vr_big_endian',
          transfer_syntax_uid='1.2.840.10008.1.2.2',
          expected_mock_get_frame_from_cache_call_count=0,
          expected_frame_bytes=50,
      ),
      dict(
          testcase_name='jpeg_baseline_process_1',
          transfer_syntax_uid='1.2.840.10008.1.2.4.50',
          expected_mock_get_frame_from_cache_call_count=0,
          expected_frame_bytes=50,
      ),
      dict(
          testcase_name='unknown_transfer_syntax',
          transfer_syntax_uid='1.2.3',
          expected_mock_get_frame_from_cache_call_count=0,
          expected_frame_bytes=50,
      ),
  ])
  def test_get_frame_server_side_use_cache_only_if_image_in_raw_transfer_syntax(
      self,
      transfer_syntax_uid,
      expected_mock_get_frame_from_cache_call_count,
      expected_frame_bytes,
  ):
    with mock.patch.object(
        dicom_slide.DicomSlide,
        '_get_frame_bytes_from_server',
        return_value=b'2',
    ) as mock_get_frame_from_server:
      with mock.patch.object(
          dicom_slide.DicomSlide,
          '_get_cached_frame_bytes',
          return_value=b'1',
      ) as mock_get_frame_from_cache:
        slide = dicom_slide.DicomSlide(
            self.mock_dwi,
            self.dicom_series_path,
            enable_client_slide_frame_decompression=True,
        )
        _init_test_slide_level_map(
            slide, 6, 6, 1, 1, 0, 8, 1, transfer_syntax_uid
        )
        level = slide.get_level_by_magnification(slide.native_magnification)
        result = slide._get_frame_server_transcoding(
            level, level.instances[0], 1  # pytype: disable=attribute-error
        )
        self.assertEqual(
            mock_get_frame_from_cache.call_count,
            expected_mock_get_frame_from_cache_call_count,
        )
        self.assertEqual(
            mock_get_frame_from_server.call_count,
            1 - expected_mock_get_frame_from_cache_call_count,
        )
    self.assertEqual(result.tolist(), [[[expected_frame_bytes]]])

  def test_get_frame_from_cache(self):
    self.mock_dwi.get_frame_image.return_value = b'abc123abc123'
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Set frame size to 2x2 for the first level.
    level_1 = slide._level_map._level_map[1]
    _init_test_slide_level_map(
        slide,
        level_1.width,
        level_1.height,
        2,
        2,
        level_1.frame_number_min,
        level_1.frame_number_max,
        level_1.samples_per_pixel,
        level_1.transfer_syntax_uid,
    )
    frame_1 = slide.get_frame(slide.native_magnification, 1)
    frame_2 = slide.get_frame(slide.native_magnification, 1)
    self.assertIsNotNone(frame_1)
    self.assertIsNotNone(frame_2)
    self.assertTrue(
        np.array_equal(frame_1.image_np, frame_2.image_np),
        'Cached frame not equal to original frame.',
    )
    self.mock_dwi.get_frame_image.assert_called_once()

  def test_get_patch_with_invalid_magnification_raise_error(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    with self.assertRaises(ez_wsi_errors.MagnificationLevelNotFoundError):
      slide.get_patch(magnification.Magnification.FromString('80X'), 0, 0, 1, 1)

  @parameterized.parameters(
      (-4, -4, 4, 4),
      (0, 7, 4, 4),
      (6, 7, 1, 2),
  )
  def test_get_patch_with_out_of_scope_patch_raise_error(
      self, x: int, y: int, width: int, height: int
  ):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Create an image at the native level, with the following properties:
    # frame size = 2 x 2
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values, with frame boundaries, are shown below:
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1'
    )
    with self.assertRaises(ez_wsi_errors.SectionOutOfImageBoundsError):
      slide.get_patch(
          slide.native_magnification, x, y, width, height
      ).image_bytes()

  @parameterized.named_parameters(
      dict(
          testcase_name='align_with_frame_boundary',
          x=0,
          y=0,
          width=4,
          height=4,
          expected_array=[
              0,
              1,
              4,
              5,
              2,
              3,
              6,
              7,
              12,
              13,
              16,
              17,
              14,
              15,
              18,
              19,
          ],
      ),
      dict(
          testcase_name='across_more_than_2_frames_in_both_x_and_y_direction',
          x=1,
          y=1,
          width=4,
          height=4,
          expected_array=[
              3,
              6,
              7,
              10,
              13,
              16,
              17,
              20,
              15,
              18,
              19,
              22,
              25,
              28,
              29,
              32,
          ],
      ),
      dict(
          testcase_name='single_frame',
          x=2,
          y=2,
          width=2,
          height=2,
          expected_array=[16, 17, 18, 19],
      ),
      dict(
          testcase_name='single_pixel',
          x=3,
          y=3,
          width=1,
          height=1,
          expected_array=[19],
      ),
      dict(
          testcase_name='single_row_across_multiple_frames',
          x=1,
          y=4,
          width=4,
          height=1,
          expected_array=[25, 28, 29, 32],
      ),
      dict(
          testcase_name='extends_beyond_the_width_of_the_image',
          x=5,
          y=1,
          width=2,
          height=3,
          expected_array=[11, 0, 21, 0, 23, 0],
      ),
      dict(
          testcase_name='extends_beyond_the_height_of_the_image',
          x=0,
          y=5,
          width=4,
          height=2,
          expected_array=[26, 27, 30, 31, 0, 0, 0, 0],
      ),
      dict(
          testcase_name=(
              'extends_beyond_both_the_width_and_the_height_of_the_image'
          ),
          x=4,
          y=4,
          width=3,
          height=3,
          expected_array=[32, 33, 0, 34, 35, 0, 0, 0, 0],
      ),
      dict(
          testcase_name='starts_outside_the_scope_of_the_image',
          x=-1,
          y=-1,
          width=3,
          height=3,
          expected_array=[0, 0, 0, 0, 0, 1, 0, 2, 3],
      ),
      dict(
          testcase_name=(
              'starts_outside_the_scope_of_the_image_and_extends_beyond_the_'
              'scope'
          ),
          x=-1,
          y=4,
          width=8,
          height=3,
          expected_array=[
              0,
              24,
              25,
              28,
              29,
              32,
              33,
              0,
              0,
              26,
              27,
              30,
              31,
              34,
              35,
              0,
              0,
              0,
              0,
              0,
              0,
              0,
              0,
              0,
          ],
      ),
  )
  def test_get_patch_with_valid_input(
      self, x: int, y: int, width: int, height: int, expected_array: List[int]
  ):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Create an image at the native level, with the following properties:
    # frame size = 2 x 2
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values are generated on-the-fly by fake_get_frame_image().
    # The actual pixel values, with frame boundaries, are shown below:
    #
    # --------+---------+-------
    #  0   1  |  4   5  |  8  9
    #  2   3  |  6   7  | 10 11
    # --------+---------+-------
    # 12  13  | 16  17  | 20 21
    # 14  15  | 18  19  | 22 23
    # --------+---------+-------
    # 24  25  | 28  29  | 32 33
    # 26  27  | 30  31  | 34 35
    # --------+---------+-------
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1'
    )
    self.mock_dwi.get_frame_image.side_effect = _fake_get_frame_raw_image
    patch = slide.get_patch(slide.native_magnification, x, y, width, height)
    self.assertEqual(patch.magnification, slide.native_magnification)
    self.assertEqual(
        [patch.x, patch.y, patch.width, patch.height], [x, y, width, height]
    )
    self.assertEqual(
        np.asarray(expected_array, np.uint8).reshape(height, width, 1).tolist(),
        patch.image_bytes().tolist(),
    )

  @parameterized.named_parameters([
      dict(
          testcase_name='all_frames_described_at_least_once',
          patch_pos_dim_list=[(0, 0, 4, 4), (-1, 4, 8, 3), (-1, -1, 3, 3)],
          expected_array=[1, 2, 4, 5, 7, 8, 9],
      ),
      dict(
          testcase_name='a_single_patch_bounds_region_describes_all_frames',
          patch_pos_dim_list=[(1, 1, 4, 4)],
          expected_array=[1, 2, 3, 4, 5, 6, 7, 8, 9],
      ),
      dict(
          testcase_name='patch_bounds_describe_frames_[5]_[5]_[7,8,9]',
          patch_pos_dim_list=[(2, 2, 2, 2), (3, 3, 1, 1), (1, 4, 4, 1)],
          expected_array=[5, 7, 8, 9],
      ),
      dict(
          testcase_name=(
              'patches_describes_subset_of_image_and_extend_beyond_image_bounds'
          ),
          patch_pos_dim_list=[
              (4, 4, 3, 3),
              (0, 5, 4, 2),
              (0, 5, 4, 2),
              (5, 1, 2, 3),
          ],
          expected_array=[3, 6, 7, 8, 9],
      ),
      dict(
          testcase_name='patches_describes_subset_of_image_frames_with_overlap',
          patch_pos_dim_list=[(-1, 4, 8, 3), (-1, -1, 3, 3)],
          expected_array=[1, 7, 8, 9],
      ),
  ])
  def test_get_patch_dicom_instances_and_frame_indexes(
      self,
      patch_pos_dim_list: List[Tuple[int, int, int, int]],
      expected_array: List[int],
  ):
    # Tests takes list of frame patch bounding regions and tests that the
    # checks that the returned frame numbers match those in the list.
    # Some of the tested patch bounding regions overlap (i.e. would share common
    # frames. The list retruned should be sorted and not have duplicates.

    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Create an image at the native level, with the following properties:
    # frame size = 2 x 2
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values are generated on-the-fly by fake_get_frame_image().
    # The actual pixel values, with frame boundaries, are shown below:
    #
    # --------+---------+-------
    #  0   1  |  4   5  |  8  9
    #  2   3  |  6   7  | 10 11
    # --------+---------+-------
    # 12  13  | 16  17  | 20 21
    # 14  15  | 18  19  | 22 23
    # --------+---------+-------
    # 24  25  | 28  29  | 32 33
    # 26  27  | 30  31  | 34 35
    # --------+---------+-------
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1', mock_path=True
    )
    self.mock_dwi.get_frame_image.side_effect = _fake_get_frame_raw_image
    expected = {
        str(dicom_path.FromPath(slide.path, instance_uid='1')): expected_array
    }
    patch_list = [
        slide.get_patch(slide.native_magnification, x, y, width, height)
        for x, y, width, height in patch_pos_dim_list
    ]
    instance_frame_map = slide.get_patch_bounds_dicom_instance_frame_numbers(
        patch_list[0].magnification,
        [patch.patch_bounds for patch in patch_list],
    )

    self.assertEqual(instance_frame_map, expected)

  def test_get_dicom_instance_frames_across_concat_instances(self):
    mock_dwi = dicom_test_utils.create_mock_dicom_web_interface(
        dicom_test_utils.INSTANCE_CONCATENATION_TEST_DATA_PATH
    )
    self.dicom_series_path = dicom_path.FromString(
        dicom_test_utils.TEST_DICOM_SERIES
    )
    slide = dicom_slide.DicomSlide(
        mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    level_1 = slide._level_map._level_map[1]
    _init_test_slide_level_map(
        slide,
        level_1.width,
        level_1.height,
        level_1.frame_width,
        level_1.frame_height,
        level_1.frame_number_min,
        level_1.frame_number_max,
        1,
        level_1.transfer_syntax_uid,
        mock_path=True,
    )
    mock_dwi.get_frame_image.side_effect = _fake_get_frame_raw_image
    patch_list = [slide.get_patch(slide.native_magnification, 0, 0, 8, 6)]
    instance_frame_map = slide.get_patch_bounds_dicom_instance_frame_numbers(
        patch_list[0].magnification,
        [patch.patch_bounds for patch in patch_list],
    )
    expected = {
        str(dicom_path.FromPath(slide.path, instance_uid='1')): [
            1,
            2,
            3,
            4,
            5,
            6,
        ],
        str(dicom_path.FromPath(slide.path, instance_uid='2')): [
            1,
            2,
            3,
            4,
            5,
            6,
        ],
    }
    self.assertEqual(instance_frame_map, expected)

  def test_get_dicom_instance_invalid_mag(self):
    mock_dwi = dicom_test_utils.create_mock_dicom_web_interface(
        dicom_test_utils.INSTANCE_CONCATENATION_TEST_DATA_PATH
    )
    self.dicom_series_path = dicom_path.FromString(
        dicom_test_utils.TEST_DICOM_SERIES
    )
    slide = dicom_slide.DicomSlide(
        mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    level_1 = slide._level_map._level_map[1]
    _init_test_slide_level_map(
        slide,
        level_1.width,
        level_1.height,
        level_1.frame_width,
        level_1.frame_height,
        level_1.frame_number_min,
        level_1.frame_number_max,
        1,
        level_1.transfer_syntax_uid,
        mock_path=True,
    )
    mock_dwi.get_frame_image.side_effect = _fake_get_frame_raw_image
    patch_list = [slide.get_patch(slide.native_magnification, 0, 0, 8, 6)]
    instance_frame_map = slide.get_patch_bounds_dicom_instance_frame_numbers(
        patch_list[0].magnification.next_higher_magnification,
        [patch.patch_bounds for patch in patch_list],
    )
    self.assertEmpty(instance_frame_map)

  @parameterized.named_parameters(
      dict(
          testcase_name='align_with_frame_boundary',
          x=0,
          y=0,
          width=4,
          height=4,
          expected_array=[0, 1, 3, 4],
      ),
      dict(
          testcase_name='across_more_than_2_frames_in_both_x_and_y_direction',
          x=1,
          y=1,
          width=4,
          height=4,
          expected_array=[0, 1, 2, 3, 4, 5, 6, 7, 8],
      ),
      dict(
          testcase_name='single_frame',
          x=2,
          y=2,
          width=2,
          height=2,
          expected_array=[4],
      ),
      dict(
          testcase_name='single_ pixel',
          x=3,
          y=3,
          width=1,
          height=1,
          expected_array=[4],
      ),
      dict(
          testcase_name='single_row_across_multiple_frames',
          x=1,
          y=4,
          width=4,
          height=1,
          expected_array=[6, 7, 8],
      ),
      dict(
          testcase_name='extends_beyond_the_width_of_the_image',
          x=5,
          y=1,
          width=2,
          height=3,
          expected_array=[2, 5],
      ),
      dict(
          testcase_name='extends_beyond_the_height_of_the_image',
          x=0,
          y=5,
          width=4,
          height=2,
          expected_array=[6, 7],
      ),
      dict(
          testcase_name=(
              'extends_beyond_both_the_width_and_the_height_of_the_image'
          ),
          x=4,
          y=4,
          width=3,
          height=3,
          expected_array=[8],
      ),
      dict(
          testcase_name='starts_outside_the_scope_of_the_image',
          x=-1,
          y=-1,
          width=3,
          height=3,
          expected_array=[0],
      ),
      dict(
          testcase_name=(
              'Starts outside the scope of the image, and extends beyond the'
              ' scope.'
          ),
          x=-1,
          y=4,
          width=8,
          height=3,
          expected_array=[6, 7, 8],
      ),
  )
  def test_get_patch_frame_numbers_with_valid_input(
      self, x: int, y: int, width: int, height: int, expected_array: List[int]
  ):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Create an image at the native level, with the following properties:
    # frame size = 2 x 2
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values are generated on-the-fly by fake_get_frame_image().
    # The actual pixel values, with frame boundaries, are shown below:
    #
    # ----+----+-----
    #  1  |  2 |  3
    # ----+----+-----
    #  4  |  5 |  6
    # ----+----+-----
    #  7  |  8 |  9
    # ----+----+-----
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1'
    )
    self.mock_dwi.get_frame_image.side_effect = _fake_get_frame_raw_image
    patch = slide.get_patch(slide.native_magnification, x, y, width, height)
    self.assertEqual(patch.magnification, slide.native_magnification)
    self.assertEqual(
        [patch.x, patch.y, patch.width, patch.height], [x, y, width, height]
    )
    # sort order is expected
    self.assertEqual(expected_array, list(patch.frame_number()))

  def test_get_image(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Create an image at the native level, with the following properties:
    # frame size = 2 x 2
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values are generated on-the-fly by fake_get_frame_image().
    # The actual pixel values, with frame boundaries, are shown below:
    #
    # --------+---------+-------
    #  0   1  |  4   5  |  8  9
    #  2   3  |  6   7  | 10 11
    # --------+---------+-------
    # 12  13  | 16  17  | 20 21
    # 14  15  | 18  19  | 22 23
    # --------+---------+-------
    # 24  25  | 28  29  | 32 33
    # 26  27  | 30  31  | 34 35
    # --------+---------+-------
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1'
    )
    self.mock_dwi.get_frame_image.side_effect = _fake_get_frame_raw_image
    image = slide.get_image(slide.native_magnification)
    self.assertEqual(image.magnification, slide.native_magnification)
    self.assertEqual([image.width, image.height], [6, 6])
    self.assertEqual(
        np.asarray(
            [
                0,
                1,
                4,
                5,
                8,
                9,
                2,
                3,
                6,
                7,
                10,
                11,
                12,
                13,
                16,
                17,
                20,
                21,
                14,
                15,
                18,
                19,
                22,
                23,
                24,
                25,
                28,
                29,
                32,
                33,
                26,
                27,
                30,
                31,
                34,
                35,
            ],
            np.uint8,
        )
        .reshape(6, 6, 1)
        .tolist(),
        image.image_bytes().tolist(),
    )
    self.assertEqual(self.mock_dwi.get_frame_image.call_count, 9)

  def test_set_dicom_slide_frame_cache(self):
    mock_frame_cache = typing.cast(
        abstract_slide_frame_cache.AbstractSlideFrameCache, MockFrameCache()
    )
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    slide.slide_frame_cache = mock_frame_cache
    self.assertIs(slide.slide_frame_cache, mock_frame_cache)

  def test_get_image_from_slide_frame_cache_by_instance(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    slide.slide_frame_cache = typing.cast(
        abstract_slide_frame_cache.AbstractSlideFrameCache, MockFrameCache()
    )
    level = slide._level_map._level_map[1]
    self.assertEqual(
        slide._get_cached_frame_bytes(level.instances[0], 1),
        b'\x00\x01\x02\x03',
    )

  def test_get_image_from_slide_frame_cache_by_path(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    slide.slide_frame_cache = typing.cast(
        abstract_slide_frame_cache.AbstractSlideFrameCache, MockFrameCache()
    )
    level = slide._level_map._level_map[1]
    self.assertEqual(
        slide._get_cached_frame_bytes(level.instances[0], 2),
        b'\x04\x05\x06\x07',
    )

  def test_get_image_from_slide_frame_cache(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    slide.slide_frame_cache = typing.cast(
        abstract_slide_frame_cache.AbstractSlideFrameCache, MockFrameCache()
    )
    # Create an image at the native level, with the following properties:
    # frame size = 2 x 2
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values are generated on-the-fly by fake_get_frame_image().
    # The actual pixel values, with frame boundaries, are shown below:
    #
    # --------+---------+-------
    #  0   1  |  4   5  |  8  9
    #  2   3  |  6   7  | 10 11
    # --------+---------+-------
    # 12  13  | 16  17  | 20 21
    # 14  15  | 18  19  | 22 23
    # --------+---------+-------
    # 24  25  | 28  29  | 32 33
    # 26  27  | 30  31  | 34 35
    # --------+---------+-------
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1'
    )
    image = slide.get_image(slide.native_magnification)
    self.assertEqual(image.magnification, slide.native_magnification)
    self.assertEqual([image.width, image.height], [6, 6])
    self.assertEqual(
        np.asarray(
            [
                0,
                1,
                4,
                5,
                8,
                9,
                2,
                3,
                6,
                7,
                10,
                11,
                12,
                13,
                16,
                17,
                20,
                21,
                14,
                15,
                18,
                19,
                22,
                23,
                24,
                25,
                28,
                29,
                32,
                33,
                26,
                27,
                30,
                31,
                34,
                35,
            ],
            np.uint8,
        )
        .reshape(6, 6, 1)
        .tolist(),
        image.image_bytes().tolist(),
    )
    self.assertEqual(self.mock_dwi.get_frame_image.call_count, 0)

  @parameterized.parameters((3, 3, 2, 2), (4, 5, 3, 3), (-1, -1, 1, 1))
  def test_copy_overlapped_region_with_invalid_input_raise_error(
      self, dst_x: int, dst_y: int, dst_width: int, dst_height: int
  ):
    mag = magnification.Magnification.FromString('40X')
    src_frame = dicom_slide.Frame(0, 0, 3, 3, np.ndarray((3, 3, 3), np.uint8))
    dst_patch = dicom_slide.Patch(mag, dst_x, dst_y, dst_width, dst_height)
    with self.assertRaises(ez_wsi_errors.PatchIntersectionNotFoundError):
      dst_np = np.ndarray((dst_height, dst_width, 3), np.uint8)
      dst_patch._copy_overlapped_region(src_frame, dst_np)

  @parameterized.parameters(
      (1, 0, 2, 1, [[[0, 0, 0], [0, 0, 0]], [[1, 1, 1], [2, 2, 2]]]),
      (1, 1, 2, 2, [[[1, 1, 1], [2, 2, 2]], [[4, 4, 4], [5, 5, 5]]]),
      (3, 2, 1, 2, [[[6, 6, 6], [0, 0, 0]], [[9, 9, 9], [0, 0, 0]]]),
  )
  def test_copy_overlapped_region_with_valid(
      self,
      dst_x: int,
      dst_y: int,
      expected_region_width: int,
      expected_region_height: int,
      expected_array: np.ndarray,
  ):
    mag = magnification.Magnification.FromString('40X')
    src_frame = dicom_slide.Frame(
        1,
        1,
        3,
        3,
        np.array(
            [
                [[1, 1, 1], [2, 2, 2], [3, 3, 3]],
                [[4, 4, 4], [5, 5, 5], [6, 6, 6]],
                [[7, 7, 7], [8, 8, 8], [9, 9, 9]],
            ],
            np.uint8,
        ),
    )
    dst_patch = dicom_slide.Patch(mag, dst_x, dst_y, 2, 2)
    dst_np = np.zeros((2, 2, 3), np.uint8)

    region_width, region_height = dst_patch._copy_overlapped_region(
        src_frame, dst_np
    )

    self.assertEqual(
        (expected_region_width, expected_region_height),
        (region_width, region_height),
    )
    self.assertTrue(np.array_equal(expected_array, dst_np))

  @parameterized.parameters(
      (1, 1, 3, 3, 0, 0), (1, 1, 2, 2, 1, 1), (1, 1, 3, 3, 0, 0)
  )
  def test_copy_ndarray_with_invalid_input_raise_error(
      self,
      src_x: int,
      src_y: int,
      width: int,
      height: int,
      dst_x: int,
      dst_y: int,
  ):
    src_array = np.ndarray((3, 3, 3), np.uint8)
    dst_array = np.ndarray((2, 2, 3), np.uint8)
    with self.assertRaises(ez_wsi_errors.SectionOutOfImageBoundsError):
      dicom_slide._copy_ndarray(
          src_array, src_x, src_y, width, height, dst_array, dst_x, dst_y
      )

  @parameterized.parameters(
      (1, 1, 2, 2, 0, 0, [[[5, 5, 5], [6, 6, 6]], [[8, 8, 8], [9, 9, 9]]]),
      (1, 1, 1, 2, 1, 0, [[[0, 0, 0], [5, 5, 5]], [[0, 0, 0], [8, 8, 8]]]),
  )
  def test_copy_ndarray_with_valid_input(
      self,
      src_x: int,
      src_y: int,
      width: int,
      height: int,
      dst_x: int,
      dst_y: int,
      expected_array: np.ndarray,
  ):
    src_array = np.array(
        [
            [[1, 1, 1], [2, 2, 2], [3, 3, 3]],
            [[4, 4, 4], [5, 5, 5], [6, 6, 6]],
            [[7, 7, 7], [8, 8, 8], [9, 9, 9]],
        ],
        np.uint8,
    )
    dst_array = np.zeros((2, 2, 3), np.uint8)
    dicom_slide._copy_ndarray(
        src_array, src_x, src_y, width, height, dst_array, dst_x, dst_y
    )
    self.assertTrue(np.array_equal(expected_array, dst_array))

  def test_get_slide_patch_id(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        accession_number='slideid',
        enable_client_slide_frame_decompression=True,
    )
    self.assertEqual('slideid', slide.accession_number)
    patch = slide.get_patch(
        magnification.Magnification.FromString('10X'),
        x=10,
        y=20,
        width=100,
        height=200,
    )
    self.assertEqual('slideid:M_10X:000100x000200+000010+000020', patch.id)

  def test_get_patch_from_jpeg(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Create an image at the native level, with the following properties:
    # frame size = 156 x 455
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values are generated on-the-fly by fake_get_frame_image().
    # The actual pixel values, with frame boundaries, are shown below:
    #
    # ---+----+----
    #  0 |  2 |  3
    # ---+----+----
    #  4 |  5 |  6
    # ---+----+----
    #  7 |  8 |  9
    # ---+----+----
    _init_test_slide_level_map(
        slide, 1365, 468, 455, 156, 0, 3, 3, '1.2.840.10008.1.2.4.50'
    )
    self.mock_dwi.get_frame_image.side_effect = _fake_get_frame_image_jpeg
    x = 100
    y = 50
    width = 300
    height = 100
    expected_array = np.asarray(PIL.Image.open(dicom_test_utils.TEST_JPEG_PATH))
    expected_array = expected_array[y : y + height, x : x + width, :]

    patch = slide.get_patch(slide.native_magnification, x, y, width, height)
    self.assertEqual(patch.magnification, slide.native_magnification)
    self.assertEqual(
        [patch.x, patch.y, patch.width, patch.height], [x, y, width, height]
    )
    self.assertTrue(np.array_equal(expected_array, patch.image_bytes()))
    self.mock_dwi.get_frame_image.assert_called_once()

  def test_get_patch_from_jpeg_locally_fails_over_server_side_decoding(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    # Create an image at the native level, with the following properties:
    # frame size = 156 x 455
    # image size = 6 x 6
    # frame numbers: 0~8
    # samples per pixel = 1
    # The image pixel values are generated on-the-fly by fake_get_frame_image().
    # The actual pixel values, with frame boundaries, are shown below:
    #
    # --------+---------+-------
    #  0   1  |  4   5  |  8  9
    #  2   3  |  6   7  | 10 11
    # --------+---------+-------
    # 12  13  | 16  17  | 20 21
    # 14  15  | 18  19  | 22 23
    # --------+---------+-------
    # 24  25  | 28  29  | 32 33
    # 26  27  | 30  31  | 34 35
    # --------+---------+-------
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1'
    )
    self.mock_dwi.get_frame_image.side_effect = _fake_get_frame_raw_image
    x = 1
    y = 2
    width = 4
    height = 2
    expected_array = np.asarray(
        [13, 16, 17, 20, 15, 18, 19, 22], np.uint8
    ).reshape(height, width, 1)

    patch = slide.get_patch(slide.native_magnification, x, y, width, height)
    self.assertEqual(patch.magnification, slide.native_magnification)
    self.assertEqual(
        [patch.x, patch.y, patch.width, patch.height], [x, y, width, height]
    )
    self.assertTrue(np.array_equal(expected_array, patch.image_bytes()))
    self.assertEqual(self.mock_dwi.get_frame_image.call_count, 3)

  def test_slide_levels(self):
    slide = dicom_slide.DicomSlide(
        self.mock_dwi,
        self.dicom_series_path,
        enable_client_slide_frame_decompression=True,
    )
    _init_test_slide_level_map(
        slide, 6, 6, 2, 2, 0, 8, 1, '1.2.840.10008.1.2.1'
    )
    self.assertIsInstance(slide.levels, Iterator)
    self.assertLen(list(slide.levels), 10)


if __name__ == '__main__':
  absltest.main()



# !/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Install script for ez-wsi-dicomweb."""

import setuptools

setuptools.setup(
    name='ez_wsi_dicomweb',
    version='1.0.2',
    url='https://github.com/GoogleCloudPlatform/ez-wsi-dicomweb',
    author='Google LLC.',
    author_email='no-reply@google.com',
    license='Apache 2.0',
    description=(
        'A library that provides the ability to extract an image patch from a'
        ' pathology DICOM whole slide image.'
    ),
    install_requires=[
        'numpy',
        'absl-py',
        'cachetools',
        'opencv-python',
        'mock',
        'pillow',
        (
            'hcls_imaging_ml_toolkit @'
            ' git+https://github.com/GoogleCloudPlatform/healthcare.git#subdirectory=imaging/ml/toolkit'
        ),
    ],
    packages=setuptools.find_packages(),
)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for patch_generator."""

import os
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from ez_wsi_dicomweb import dicom_slide
from ez_wsi_dicomweb import dicom_test_utils
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import magnification as mag_lib
from ez_wsi_dicomweb import patch_generator as patch_generator_lib
import numpy as np
from PIL import Image

from hcls_imaging_ml_toolkit import dicom_path


class PatchGeneratorTest(parameterized.TestCase):

  def test_iterate_tissue_mask(self):
    patch_arr = np.array(
        Image.open(
            os.path.join(
                dicom_test_utils.TEST_DATA_PATH, 'low_res_slide_img.png'
            )
        )
    )
    mock_slide = mock.create_autospec(dicom_slide.DicomSlide)
    mock_image = mock.create_autospec(dicom_slide.Image)
    mock_slide.get_image.return_value = mock_image
    mock_image.image_bytes.return_value = patch_arr
    expected = np.array(
        Image.open(
            os.path.join(
                dicom_test_utils.TEST_DATA_PATH, 'golden_inference_mask.png'
            )
        )
    )
    patch_generator = patch_generator_lib.PatchGenerator(
        mock_slide,
        mag_lib.Magnification.FromString('40X'),
        stride_size=128,
        patch_size=patch_arr.shape[0],
        max_luminance=0.8,
    )
    np.testing.assert_array_almost_equal(
        expected,
        patch_generator._normalized_tissue_mask().astype(np.uint8) * 255,
    )

    patch1 = mock.create_autospec(dicom_slide.Patch)
    patch2 = mock.create_autospec(dicom_slide.Patch)
    mock_slide.get_patch.side_effect = [patch1, patch2]
    it = iter(patch_generator)
    self.assertEqual(patch1, next(it))
    self.assertEqual(patch2, next(it))

  def test_total_num_patches(self):
    # 4x8 is white and should be filtered; 8x8 is tissue and should be kept
    patch_arr = np.concatenate(
        (np.ones((4, 8, 3)), np.ones((8, 8, 3)) * 0.1), axis=0
    )
    mock_slide = mock.create_autospec(dicom_slide.DicomSlide)
    mock_image = mock.create_autospec(dicom_slide.Image)
    mock_slide.get_image.return_value = mock_image
    mock_image.image_bytes.return_value = patch_arr
    patch_generator = patch_generator_lib.PatchGenerator(
        mock_slide,
        mag_lib.Magnification.FromString('40X'),
        stride_size=128,
        patch_size=patch_arr.shape[0],
        max_luminance=0.8,
    )

    # stride size 128 at 40X is 4 at 1.25X. Expected number of tissue patches is
    # 8 * 8 / 4 / 4 = 4
    self.assertEqual(4, patch_generator.total_num_patches())

  def test_total_num_patches_custom_magnification(self):
    # 4x8 is white and should be filtered; 8x8 is tissue and should be kept
    patch_arr = np.concatenate(
        (np.ones((4, 8, 3)), np.ones((8, 8, 3)) * 0.1), axis=0
    )

    mock_slide = mock.create_autospec(dicom_slide.DicomSlide)
    mock_image = mock.create_autospec(dicom_slide.Image)
    mock_slide.get_image.return_value = mock_image
    mock_image.image_bytes.return_value = patch_arr
    patch_generator = patch_generator_lib.PatchGenerator(
        mock_slide,
        mag_lib.Magnification.FromString('40X'),
        stride_size=128,
        patch_size=patch_arr.shape[0],
        max_luminance=0.8,
        tissue_mask_magnification=mag_lib.Magnification.FromString('2.5X'),
    )

    # stride size 128 at 40X is 8 at 2.5X. Expected number of tissue patches is
    # 8 * 8 / 8 / 8 = 1
    self.assertEqual(1, patch_generator.total_num_patches())

  def test_total_strides(self):
    patch_arr = np.ones((8, 8, 3)) * 0.1
    mock_slide = mock.create_autospec(dicom_slide.DicomSlide)
    mock_image = mock.create_autospec(dicom_slide.Image)
    mock_slide.get_image.return_value = mock_image
    mock_image.image_bytes.return_value = patch_arr
    patch_generator = patch_generator_lib.PatchGenerator(
        mock_slide,
        mag_lib.Magnification.FromString('40X'),
        stride_size=128,
        patch_size=patch_arr.shape[0],
        max_luminance=0.8,
    )

    # stride size 128 at 40X is 4 at 1.25X. Expected y_strides is 8 / 4 = 2,
    # expected x_strides is 8 / 4 = 2.
    self.assertEqual(
        patch_generator_lib.StrideCoordinate(y_strides=2, x_strides=2),
        patch_generator.total_strides(),
    )

  def test_total_strides_large_stride_size(self):
    patch_arr = np.ones((8, 8, 3)) * 0.1
    mock_slide = mock.create_autospec(dicom_slide.DicomSlide)
    mock_image = mock.create_autospec(dicom_slide.Image)
    mock_slide.get_image.return_value = mock_image
    mock_image.image_bytes.return_value = patch_arr
    patch_generator = patch_generator_lib.PatchGenerator(
        mock_slide,
        mag_lib.Magnification.FromString('40X'),
        stride_size=512,
        patch_size=patch_arr.shape[0],
        max_luminance=0.8,
    )

    # stride size 512 at 40X is 16 at 1.25X. Expected y_strides is
    # max(8 / 16, 1) = 1, expected x_strides is max(8 / 16, 1) = 1.
    self.assertEqual(
        patch_generator_lib.StrideCoordinate(y_strides=1, x_strides=1),
        patch_generator.total_strides(),
    )

  def test_empty_tissue_mask(self):
    patch_arr = np.array(
        Image.open(
            os.path.join(
                dicom_test_utils.TEST_DATA_PATH, 'low_res_slide_img.png'
            )
        )
    )
    mock_slide = mock.create_autospec(dicom_slide.DicomSlide)
    mock_image = mock.create_autospec(dicom_slide.Image)
    mock_slide.get_image.return_value = mock_image
    mock_image.image_bytes.return_value = patch_arr
    slide_uid = 'slide_uid'
    series_uid = 'series_uid'
    dicom_full_path = (
        'projects/project_name/locations/us-west1/datasets/'
        'dataset_name/dicomStores/dicom_store_name/dicomWeb/'
        f'studies/{slide_uid}/series/{series_uid}'
    )
    mock_slide.path = dicom_path.FromString(dicom_full_path)

    threshold = 0
    patch_generator = patch_generator_lib.PatchGenerator(
        mock_slide,
        mag_lib.Magnification.FromString('40X'),
        stride_size=128,
        patch_size=patch_arr.shape[0],
        max_luminance=threshold,
    )
    with self.assertRaisesRegex(
        ez_wsi_errors.DicomImageMissingRegionError,
        (
            f'Slide with study_uid {slide_uid} series_uid {series_uid} has no '
            f'region with luminance value less than threshold {threshold}.'
        ),
    ):
      next(iter(patch_generator))

  @parameterized.named_parameters(
      dict(
          testcase_name='base_case',
          y_strides=0,
          x_strides=0,
          patch_size=128,
          stride_size=128,
          expected_patch_bounds=dicom_slide.PatchBounds(
              x_origin=0, y_origin=0, width=128, height=128
          ),
      ),
      # Given that patches are originiated in the upper left corner a larger
      # patch should be offset to the left.
      dict(
          testcase_name='larger_patch',
          y_strides=0,
          x_strides=0,
          patch_size=299,
          stride_size=128,
          expected_patch_bounds=dicom_slide.PatchBounds(
              x_origin=-85, y_origin=-85, width=299, height=299
          ),
      ),
      # Increasing stride should shift patch right, down by (stride-patch)/2
      dict(
          testcase_name='larger_stride',
          y_strides=0,
          x_strides=0,
          patch_size=128,
          stride_size=299,
          expected_patch_bounds=dicom_slide.PatchBounds(
              x_origin=85, y_origin=85, width=128, height=128
          ),
      ),
      # Moving coordinates should shift by stride
      dict(
          testcase_name='offset_y_stride',
          y_strides=1,
          x_strides=0,
          patch_size=299,
          stride_size=128,
          expected_patch_bounds=dicom_slide.PatchBounds(
              x_origin=-85, y_origin=(-85 + 128), width=299, height=299
          ),
      ),
      # Moving coordinates should shift by stride
      dict(
          testcase_name='offset_x_stride',
          y_strides=0,
          x_strides=1,
          patch_size=299,
          stride_size=128,
          expected_patch_bounds=dicom_slide.PatchBounds(
              x_origin=(-85 + 128), y_origin=-85, width=299, height=299
          ),
      ),
  )
  def test_coordinates_conversion(
      self, y_strides, x_strides, patch_size, stride_size, expected_patch_bounds
  ):
    mock_slide = mock.create_autospec(dicom_slide.DicomSlide)
    patch_generator = patch_generator_lib.PatchGenerator(
        mock_slide,
        mag_lib.Magnification.FromString('40X'),
        stride_size,
        patch_size,
    )
    self.assertEqual(
        expected_patch_bounds,
        patch_generator.strides_to_patch_bounds(
            patch_generator_lib.StrideCoordinate(y_strides, x_strides)
        ),
    )
    self.assertEqual(
        patch_generator_lib.StrideCoordinate(y_strides, x_strides),
        patch_generator.patch_bounds_to_strides(expected_patch_bounds),
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Abstract class for DICOM Slide frame cache."""
import abc
import dataclasses
from typing import List, Mapping, Optional, Union

from ez_wsi_dicomweb import slide_level_map


@dataclasses.dataclass
class AbstractSlideFrameCacheStats:
  """Abstract base class for SlideFrameCacheStats."""


class AbstractSlideFrameCache(metaclass=abc.ABCMeta):
  """Abstract class for DICOM Slide frame cache."""

  @abc.abstractmethod
  def get_frame(
      self, instance_path: str, number_of_frames: int, frame_number: int
  ) -> Optional[bytes]:
    """Returns DICOM instance frame bytes or None.

    Args:
      instance_path: DICOMweb path to instance.
      number_of_frames: Total number of frames in instance.
      frame_number: Frame index within instance.  The frames are referenced in
        with 1-based indexing.

    Returns:
      Frame bytes or None.
    """

  @abc.abstractmethod
  def preload_instance_frame_numbers(
      self,
      instance_frame_numbers: Mapping[str, List[int]],
  ) -> None:
    """Preloads select instance frames from DICOM Store into cache.

    Args:
      instance_frame_numbers: Map of instance path to frame numbers.

    Returns:
      None.
    """

  @abc.abstractmethod
  def cache_whole_instance_in_memory(
      self,
      instance_paths: Union[slide_level_map.Level, slide_level_map.Instance],
      blocking: bool,
  ) -> None:
    """Caches whole DICOM instance in memory.

    Args:
      instance_paths: DICOMweb path to instance.
      blocking: Load cache as blocking operation.
    """

  @abc.abstractmethod
  def reset_cache_stats(self) -> None:
    """Resets cache status metrics."""

  @property
  @abc.abstractmethod
  def cache_stats(self) -> AbstractSlideFrameCacheStats:
    """Returns dataclass which encodes cache stats."""
